// Title:   DeformSurfel.h
// Created: Thu Sep 25 15:07:35 2003
// Authors: Richard Keiser, Mark Pauly
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __DEFORMSURFEL_H_
#define __DEFORMSURFEL_H_



#include "../../../../DataTypes/src/Vector3D.h"
#include "../../../../Core/Interfaces/src/SurfelInterface.h"
#include "../../../../Core/DataStructures/src/NeighbourHood.h"
#include "../../../../Core/Scene/src/Scene.h"
#include "../../../../Utilities/src/Matrix.h"
#include "../../../../Utilities/src/MyMathTools.h"
#include "../../../../Utilities/MagicSoft/MgcMatrix2.h"

#include "ScaleGrid.h"

#define ALLOCATE_MORE_DEFORMSURFELS 65536

typedef struct deformParams {

	bool		doTranslation;
	Vector3D	translation;
	bool		doRotation;
	Vector3D	rotationCenter;
	Vector3D	rotationAxis;
	Vector3D	rotationTangent1;
	Vector3D	rotationTangent2;
	float		angle;
} DeformParams;


/**
 * Class for the surfels of the deformable region
 *
 * @author Mark Pauly
 * @version 2.0
 * @see DeformerTool
*/
class DeformSurfel {
	
public:

	typedef enum blendingFunction {

		NONE      = 0,
		SMOOTH    = 1,
		LINEAR    = 2,
		QUADRATIC = 3,
		SMOOTH2   = 4

	} BlendingFunction;

	/**
	 * initializes this <code>DeformSurfel</code> with parameters of <code>surfel</code>
	 */
	inline void initialize (SurfelInterface *surfel);
	/**
	 * initializes this <code>DeformSurfel</code> with parameters of <code>surfel</code>
	 */	
	inline void initialize (DeformSurfel *surfel);

	/**
	 * returns the normalized distance of 'position' to the zero/one region
	 */
	inline float compNormDist (ScaleGrid *scaleGrid, Vector3D position);
	/**
	 * returns the normalized distance of 'position' and computes the undeformed position
	 */
	inline float compNormDist (ScaleGrid *scaleGrid, Vector2D position, Vector3D &newUndeformedPosition);
	/**
	 * computes the distance and tangent axes of this <code>DeformSurfel</code>
	 *
	 * @see #computeScale
	 */
	inline void compNormDists (ScaleGrid *scaleGrid);
	
	/**
	 * computes the scale for the position and tangent axes
	 *
	 * @see #compNormDists
	 */
	inline void computeScale (NeighbourHood *scaleRegion, float &maxScaleDistance);		
	/**
	 * rescale the position and tangent axes
	 *
	 * @see #computeScale
	 */
	inline void rescaleScale (float maxScaleDistance);

	
	/**
	 * deforms a single point
	 */
	static inline Vector3D deformPoint (Vector3D sourcePoint, float scaleFactor, DeformParams deformParams);

	/**
	 * deformation. computes new position, normal & tangent axes
	 */
	inline void deform (DeformParams &deformParams);

	/**
	 * computes the scale factor for the deformation according to blendingFunction
	 */
	inline void computeDeformationScales (BlendingFunction blendingFunction);

	/**
	 * a smooth blending function
	 */
	inline float smoothBlendingFunction (float distance);
	/**
	 * a smooth blending function
	 */
	inline float smoothBlendingFunction2 (float distance);
	/**
	 * a linear blending function
	 */
	inline float linearBlendingFunction (float ditance);
	/**
	 * a quadratic blending function
	 */
	inline float quadraticBlendingFunction (float distance);

	/**
	 * checks if the ellipse should be split. Currently only the ratio of the axes is considered not the size of the 
	 * ellipse. Thus ellipses that are squeezed might also be split.
	 */
	inline bool checkForSplit (float minAxisRatioSquared);
	
	/**
	 * splits a surfel. the surfel is tested using checkForSplit and two new surfels are created if the criterion is fulfilled.
	 * the new surfels are recursively tested for splitting again
	 */
	inline void splitSurfel (std::vector<DeformSurfel> *deformSurfels, unsigned int &numDeformSurfels, unsigned int deformSurfelIndex, 
		float minAxisRatioSquared, SurfelCollection* objectSurfels, BlendingFunction blendingFunction,
		ScaleGrid *scaleGrid, QRgb *color);

	/**
	 * recomputes the principal axes as a weighted sum of axisOne and axisTwo
	 */
	inline void computePrincipalAxes (Vector3D axisOne, Vector3D axisTwo, Vector2D &principalAxisOne, Vector2D &principalAxisTwo);

	/**
	 * sets this <code>DeformSurfel</code> back to its original position & tangent axes
	 */
	inline void undoDeformation();

	/**
	 * returns the surfel which is represented by this <code>DeformSurfel</code>
	 */
	inline SurfelInterface* getSurfel();
	/** 
	 * returns the position scale
	*/
	inline float getScalePosition();


private:

	SurfelInterface *surfel;			// pointer to surfel in object's surfelCollection
	
	float           distancePosition,   // normalized distance of surfel position
		            distanceAxisOne,	// normalized distance of surfel axis one
		            distanceAxisTwo;	// normalized distance of surfel axis two
	
	Vector3D        undeformedPosition, // stores the position of the surfel prior to deformation 
		            undeformedAxisOne,
					undeformedAxisTwo;

	float           scalePosition,      // additional scaling factors
		            scaleAxisOne,
		            scaleAxisTwo;

};

inline SurfelInterface* DeformSurfel::getSurfel() {
	return surfel;
}

inline float DeformSurfel::getScalePosition() {
	return scalePosition;
}

inline void DeformSurfel::initialize (DeformSurfel *deformSurfel) {

	surfel				= deformSurfel->surfel;			// pointer to surfel in object's surfelCollection

	distancePosition	= deformSurfel->distancePosition;   // normalized distance of surfel position
	distanceAxisOne		= deformSurfel->distanceAxisOne;	// normalized distance of surfel axis one
	distanceAxisTwo		= deformSurfel->distanceAxisTwo;	// normalized distance of surfel axis two
	
	undeformedPosition	= deformSurfel->undeformedPosition; // stores the position of the surfel prior to deformation 
	undeformedAxisOne	= deformSurfel->undeformedAxisOne;
	undeformedAxisTwo	= deformSurfel->undeformedAxisTwo;

	scalePosition		= deformSurfel->scalePosition;      // additional scaling factors
	scaleAxisOne		= deformSurfel->scaleAxisOne;
	scaleAxisTwo		= deformSurfel->scaleAxisTwo;	
}

inline void DeformSurfel::initialize (SurfelInterface *surfel) {

	this->surfel = surfel;

	this->undeformedPosition = surfel->getPosition();
	this->undeformedAxisOne  = surfel->getAxisOne();
	this->undeformedAxisTwo  = surfel->getAxisTwo();
}


inline float DeformSurfel::compNormDist (ScaleGrid *scaleGrid,
										 Vector3D position) {

	return scaleGrid->getNormalizedDistance (position);
}



inline float DeformSurfel::compNormDist (ScaleGrid *scaleGrid,
										 Vector2D position,
										 Vector3D &newUndeformedPosition) {

	newUndeformedPosition = this->undeformedPosition + position[0] * this->undeformedAxisOne + position[1] * this->undeformedAxisTwo;
	return this->compNormDist (scaleGrid, newUndeformedPosition);
}



inline void DeformSurfel::compNormDists (ScaleGrid *scaleGrid) {

	this->distancePosition = compNormDist (scaleGrid, this->undeformedPosition);
	this->distanceAxisOne  = compNormDist (scaleGrid, this->undeformedPosition + this->undeformedAxisOne);
	this->distanceAxisTwo  = compNormDist (scaleGrid, this->undeformedPosition + this->undeformedAxisTwo);
}

inline void DeformSurfel::computeScale (NeighbourHood *scaleRegion, float &maxScaleDistance) {

	float scaleDistance;
	unsigned int j;
	unsigned int numScaleNeighbors = scaleRegion->getNofNeighbours();
		
	// position 
	scaleRegion->setSourcePoint (this->undeformedPosition);
	scaleDistance = 0;
	for (j = 0; j < numScaleNeighbors; j++) {		
		scaleDistance += sqrt (scaleRegion->getSquaredDistance (j));
	}
	this->scalePosition *= scaleDistance;
	if (scaleDistance > maxScaleDistance) {
		maxScaleDistance = scaleDistance;
	}
	
	// axis one
	scaleRegion->setSourcePoint (this->undeformedPosition + this->undeformedAxisOne);
	scaleDistance = 0;
	for (j = 0; j < numScaleNeighbors; j++) {		
		scaleDistance += sqrt (scaleRegion->getSquaredDistance (j));
	}
	this->scaleAxisOne *= scaleDistance;
	if (scaleDistance > maxScaleDistance) {
		maxScaleDistance = scaleDistance;
	}
	
	// axis two
	scaleRegion->setSourcePoint (this->undeformedPosition + this->undeformedAxisTwo);
	scaleDistance = 0;
	for (j = 0; j < numScaleNeighbors; j++) {		
		scaleDistance += sqrt (scaleRegion->getSquaredDistance (j));
	}
	this->scaleAxisTwo *= scaleDistance;
	if (scaleDistance > maxScaleDistance) {
		maxScaleDistance = scaleDistance;
	}
	
}


	
inline void DeformSurfel::rescaleScale (float maxScaleDistance) {

	this->scalePosition /= maxScaleDistance;
	this->scaleAxisOne  /= maxScaleDistance;
	this->scaleAxisTwo  /= maxScaleDistance;
}			
					

inline Vector3D DeformSurfel::deformPoint (Vector3D sourcePoint,
										   float scaleFactor,
										   DeformParams deformParams) {
	
	Vector3D destPoint;
	
	if (deformParams.doRotation) {
	
		Vector2D localPoint, rotatedPoint;
		float cosPhi, sinPhi;

		deformParams.angle *= scaleFactor;

		sourcePoint -= deformParams.rotationCenter;
		
		localPoint[0] = Vector3D::dotProduct (deformParams.rotationTangent1, sourcePoint);
		localPoint[1] = Vector3D::dotProduct (deformParams.rotationTangent2, sourcePoint);

		cosPhi = cos(deformParams.angle);
		sinPhi = sin(deformParams.angle);
		
		rotatedPoint[0] = cosPhi * localPoint[0] - sinPhi * localPoint[1];
		rotatedPoint[1] = sinPhi * localPoint[0] + cosPhi * localPoint[1];

		destPoint = rotatedPoint[0] * deformParams.rotationTangent1 +
			        rotatedPoint[1] * deformParams.rotationTangent2 +
				    Vector3D::dotProduct (deformParams.rotationAxis, sourcePoint) * deformParams.rotationAxis;

		destPoint += deformParams.rotationCenter;

	}
	else {
		destPoint = sourcePoint;
	}

	if (deformParams.doTranslation) {
		destPoint += scaleFactor * deformParams.translation;
	}

	return destPoint;
}


inline void DeformSurfel::deform (DeformParams &deformParams) {
	
	Scene    *scene;
	Vector3D newPosition,
		     newAxisOne,
			 newAxisTwo,
			 newNormal;
	
	scene = Scene::getInstance();
		
	newPosition = this->deformPoint (this->surfel->getPosition(), this->scalePosition, deformParams);
	
	newAxisOne = this->deformPoint (this->surfel->getPosition() + this->surfel->getAxisOne(), 
		this->scaleAxisOne, deformParams) - newPosition;
	
	newAxisTwo = this->deformPoint (this->surfel->getPosition() + this->surfel->getAxisTwo(), 
		this->scaleAxisTwo, deformParams) - newPosition;
	
	newNormal = Vector3D::crossProduct (newAxisOne, newAxisTwo);
	newNormal.normalize();
	
	scene->setSurfelPosition    (this->surfel, newPosition, false);
	scene->setSurfelNormalTangentAxesSystem(this->surfel, newNormal, newAxisOne, newAxisTwo, false);

}




inline void DeformSurfel::computeDeformationScales (BlendingFunction blendingFunction) {

	if (blendingFunction == SMOOTH) {		
		this->scalePosition = smoothBlendingFunction (this->distancePosition);
		this->scaleAxisOne  = smoothBlendingFunction (this->distanceAxisOne);
		this->scaleAxisTwo  = smoothBlendingFunction (this->distanceAxisTwo);
	}
	else if (blendingFunction == SMOOTH2) {		
		this->scalePosition = smoothBlendingFunction2 (this->distancePosition);
		this->scaleAxisOne  = smoothBlendingFunction2 (this->distanceAxisOne);
		this->scaleAxisTwo  = smoothBlendingFunction2 (this->distanceAxisTwo);
	}

	else if (blendingFunction == LINEAR) {
		this->scalePosition = linearBlendingFunction (this->distancePosition);
		this->scaleAxisOne  = linearBlendingFunction (this->distanceAxisOne);
		this->scaleAxisTwo  = linearBlendingFunction (this->distanceAxisTwo);
	}
	else if (blendingFunction == QUADRATIC) {
		this->scalePosition = quadraticBlendingFunction (this->distancePosition);
		this->scaleAxisOne  = quadraticBlendingFunction (this->distanceAxisOne);
		this->scaleAxisTwo  = quadraticBlendingFunction (this->distanceAxisTwo);
	}
	
}



inline float DeformSurfel::smoothBlendingFunction (float distance) {

	float deformScale;
	deformScale = (distance * distance - 2 * distance);
	return deformScale * deformScale;
}

inline float DeformSurfel::smoothBlendingFunction2 (float distance) {

	float d2 = distance*distance;
	float d4 = d2*d2;

	return -43.05555556f * d4 * distance + 107.6388889f * d4 -88.11111111f * d2 * distance + 24.52777778f * d2;
}


inline float DeformSurfel::linearBlendingFunction (float distance) {
	return distance;
}


inline float DeformSurfel::quadraticBlendingFunction (float distance) {
	return distance * distance;
}




inline bool DeformSurfel::checkForSplit (float minAxisRatioSquared) {

	// compute the ratio of the sidelengths of bounding box of the parallelogram spanned by
	// the two axes. the bounding box is aligned with the longer axis.
	
	float axisOneSquaredLength = this->surfel->getAxisOne().getSquaredLength();
	float axisTwoSquaredLength = this->surfel->getAxisTwo().getSquaredLength();

	float longerLengthSquared  = MyMathTools::maxf (axisOneSquaredLength, axisTwoSquaredLength);
	float smallerLengthSquared = MyMathTools::minf (axisOneSquaredLength, axisTwoSquaredLength);


	float angle = Vector3D::dotProduct (this->surfel->getAxisOne(), this->surfel->getAxisTwo());
	float projectedLengthSquared = angle * angle / longerLengthSquared;

	if (longerLengthSquared / (smallerLengthSquared - projectedLengthSquared) > minAxisRatioSquared) {
		return true;
	}
	else {
		return false;
	}
}	



inline void DeformSurfel::splitSurfel (std::vector<DeformSurfel> *deformSurfels, 
									   unsigned int &numDeformSurfels,
									   unsigned int deformSurfelIndex,
									   float minAxisRatioSquared,
									   SurfelCollection *objectSurfels,
									   BlendingFunction blendingFunction,
									   ScaleGrid *scaleGrid,
									   QRgb *color) {

	DeformSurfel *deformSurfel,
		         *surfOne,
		         *surfTwo;
	Scene        *scene;
	Vector3D     position,
		         axisOne,
		         axisTwo;
	Vector2D	 principalAxisOne,
				 principalAxisTwo,
				 newPositionOne,
				 newPositionTwo,
				 newAxisOne,
				 newAxisTwo;
				
	Vector3D     principalAxisOne3D,
		         principalAxisTwo3D,
				 newAxisOne3D,
		         newAxisTwo3D;

	if (this->checkForSplit (minAxisRatioSquared)) {
				
		scene = Scene::getInstance();
		
		// allocate more memory if necessary
		if (deformSurfels->size() <= numDeformSurfels) {
			deformSurfels->resize (numDeformSurfels * 2);
		}
		
		surfOne = &deformSurfels->at(numDeformSurfels++);
		
		// resizing the deformSurfels array can invalidate the "this" pointer!
		// That's why we need to carry along the deformSurfelIndex, to be able to reassign the correct
		// surfel Pointer
		deformSurfel = &deformSurfels->at(deformSurfelIndex);	
		
		// surfTwo uses the same record as the surfel to be split
		// -> all changes have to be applied to surfOne first!	
		surfTwo = deformSurfel;
	
		// create a new surfel by duplicating source surfel
		surfOne->surfel = objectSurfels->addSurfel (deformSurfel->surfel, false);


		position = deformSurfel->surfel->getPosition();
		axisOne  = deformSurfel->surfel->getAxisOne();
		axisTwo  = deformSurfel->surfel->getAxisTwo();
		
		this->computePrincipalAxes (axisOne, axisTwo, principalAxisOne, principalAxisTwo);

		principalAxisOne3D = principalAxisOne[0] * axisOne + principalAxisOne[1] * axisTwo;
		principalAxisTwo3D = principalAxisTwo[0] * axisOne + principalAxisTwo[1] * axisTwo;

		// compute the positions of (position, axisOne, axisTwo) of the new surfels relative
		// to the center of surfel
		if (principalAxisOne3D.getSquaredLength() > principalAxisTwo3D.getSquaredLength()) {			
			
			newPositionOne = 0.25f  * principalAxisOne;
			newPositionTwo = -0.25f * principalAxisOne;
			newAxisOne     = 0.5f * principalAxisOne;
			newAxisTwo     = principalAxisTwo;
		}
		else {

			newPositionOne = 0.25f * principalAxisTwo;
			newPositionTwo = -0.25f * principalAxisTwo;
			newAxisOne     = principalAxisOne;
			newAxisTwo     = 0.5f * principalAxisTwo;
		}


		// need to store because surfTwo = surfel!!
		Vector3D surfTwoUndeformedPosition, 
				 surfTwoUndeformedAxisOne,
				 surfTwoUndeformedAxisTwo;

		surfOne->distancePosition   = deformSurfel->compNormDist (scaleGrid, newPositionOne, surfOne->undeformedPosition);		
		surfOne->distanceAxisOne    = deformSurfel->compNormDist (scaleGrid, newPositionOne + newAxisOne, surfOne->undeformedAxisOne);
		surfOne->undeformedAxisOne -= surfOne->undeformedPosition;
		surfOne->distanceAxisTwo    = deformSurfel->compNormDist (scaleGrid, newPositionOne + newAxisTwo, surfOne->undeformedAxisTwo);
		surfOne->undeformedAxisTwo -= surfOne->undeformedPosition;

		surfTwo->distancePosition   = deformSurfel->compNormDist (scaleGrid, newPositionTwo, surfTwoUndeformedPosition);
		surfTwo->distanceAxisOne    = deformSurfel->compNormDist (scaleGrid, newPositionTwo + newAxisOne, surfTwoUndeformedAxisOne);
		surfTwoUndeformedAxisOne -= surfTwoUndeformedPosition;		
		surfTwo->distanceAxisTwo    = deformSurfel->compNormDist (scaleGrid, newPositionTwo + newAxisTwo, surfTwoUndeformedAxisTwo);
		surfTwoUndeformedAxisTwo -= surfTwoUndeformedPosition;				



		surfTwo->undeformedPosition = surfTwoUndeformedPosition;
		surfTwo->undeformedAxisOne  = surfTwoUndeformedAxisOne;
		surfTwo->undeformedAxisTwo  = surfTwoUndeformedAxisTwo;

		scene->setSurfelPosition (surfOne->surfel, position + newPositionOne[0] * axisOne + newPositionOne[1] * axisTwo , false);
		scene->setSurfelPosition (surfTwo->surfel, position + newPositionTwo[0] * axisOne + newPositionTwo[1] * axisTwo, false);
		newAxisOne3D = newAxisOne[0] * axisOne + newAxisOne[1] * axisTwo;
		newAxisTwo3D = newAxisTwo[0] * axisOne + newAxisTwo[1] * axisTwo;

		scene->setSurfelTangentAxes (surfOne->surfel, newAxisOne3D, newAxisTwo3D, false);
		scene->setSurfelTangentAxes (surfTwo->surfel, newAxisOne3D, newAxisTwo3D, false);
		
	
		// no interpolation yet
		scene->setSurfelDiffuseColor (surfOne->surfel, deformSurfel->surfel->getDiffuseColor(), false);	
		
		// recompute deformation scales
		surfOne->computeDeformationScales (blendingFunction);
		surfTwo->computeDeformationScales (blendingFunction);

		if (color) {
			scene->setSurfelDiffuseColor(surfOne->surfel, *color, false);
			scene->setSurfelDiffuseColor(surfTwo->surfel, *color, false);
		}
	}
}



// computes the principal axes of the ellipse spanned by axisOne and axisTwo.
// (converted from Maple code)
// better solution possible??

inline void DeformSurfel::computePrincipalAxes (Vector3D axisOne, 
												Vector3D axisTwo,
												Vector2D &principalAxisOne,
												Vector2D &principalAxisTwo) {

	double t1;
	double t17;
	double t18;
	double t2;
	double t21;
	double t26;
	double t27;
	double t28;
	double t29;
	double t3;
	double t30;
	double t31;
	double t32;
	double t4;
	double t5;
	double t53;
	double t55;
	double t6;
	double phi;


	t1 = axisOne[0] * axisOne[0];
	t2 = axisTwo[0] * axisTwo[0];
	t3 = axisOne[1] * axisOne[1];
	t4 = axisTwo[1] * axisTwo[1];
	t5 = axisOne[2] * axisOne[2];
	t6 = axisTwo[2] * axisTwo[2];
	t17 = axisOne[1] * axisTwo[1];
	t18 = axisOne[2] * axisTwo[2];
	t21 = axisOne[0] * axisTwo[0];
	t26 = t1 * t1;
	t27 = t2 * t2;
	t28 = t3 * t3;
	t29 = t4 * t4;
	t30 = 2.0 * t1 * t2 + 2.0 * t1 * t3 - 2.0 * t1 * t4 + 2.0 * t1 * t5 - 2.0 * t1 * t6 + 8.0 * t17 * t18 + 8.0 * t21
		* t17 + 8.0 * t21 * t18 + t26 + t27 + t28 + t29;
	t31 = t5 * t5;
	t32 = t6 * t6;
	t53 = t31 + t32 - 2.0 * t2 * t3 + 2.0 * t2 * t4 - 2.0 * t2 * t5 + 2.0 * t2 * t6 + 2.0 * t3 * t5 - 2.0 * t3 * t6
		- 2.0 * t4 * t5 + 2.0 * t4 * t6 + 2.0 * t3 * t4 + 2.0 * t5 * t6;
	t55 = sqrt(t30 + t53);
	
	if (t21 + t17 + t18 == 00) {
		phi = 0.0f;
	}
	else {		
		phi = (atan(( - t1 + t2 - t3 + t4 - t5 + t6 + t55)/(t21 + t17 + t18)/2.0));
	}
	
	principalAxisOne[0] = cos(phi);
	principalAxisOne[1] = sin(phi);
	
	principalAxisTwo[0] = -principalAxisOne[1];
	principalAxisTwo[1] = principalAxisOne[0];	
}

inline void DeformSurfel::undoDeformation() {
	Scene *scene = Scene::getInstance();
	Vector3D normal;
	scene->setSurfelPosition(surfel, undeformedPosition, false);
	normal = Vector3D::crossProduct (undeformedAxisOne, undeformedAxisTwo);
	normal.normalize();
	scene->setSurfelNormalTangentAxesSystem(surfel, normal, undeformedAxisOne, undeformedAxisTwo, false);
}


#endif  // __DEFORMSURFEL_H_

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
