// Title:   ScaleGrid.h
// Created: Thu Sep 25 15:07:21 2003
// Authors: Richard Keiser, Mark Pauly
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __SCALEGRID_H_
#define __SCALEGRID_H_



#include "../../../../DataTypes/src/Vector3D.h"
#include "../../../../Core/Interfaces/src/SurfelInterface.h"
#include "../../../../Core/DataStructures/src/NeighbourHood.h"
#include "../../../../Core/Scene/src/Scene.h"
#include "../../../../Utilities/src/Matrix.h"
#include "../../../../Utilities/src/MyMathTools.h"



/**
 * A grid storing the scale values.
 * This grid is used by the <code>DeformationTool</code>
 *
 * @author Mark Pauly
 * @version 2.0
 */


class ScaleGrid {
	
public:
	
	/**
	 * builds a grid as the bounding box of the <code>positions</code> and computes the scale
	 * values for every grid point
	 *
	 * @param positions
	 *		  array of the position of all points
	 * @param nofPoints
	 *		  number of points in array
	 * @param nofGridPoints
	 *		  number of grid points, determines the resolution of the grid
	 * @param zeroRegion
	 *		  search datastructure for the zero region surfels
	 * @param oneRegion
	 *		  search datastructure for the one region surfels
	 * @see #computeNormalizedDistance
	 */
	ScaleGrid (Vector3D *positions, unsigned int nofPoints, unsigned int nofGridPoints, 
		NeighbourHood *zeroRegion, NeighbourHood *oneRegion);

	/**
	 * computes the scale false of a point a <code>position</code) as the smallest distance to the zeroRegion and
	 * the oneRegion as follows: value = zeroDistancePosition / (zeroDistancePosition + oneDistancePosition)
	 *
	 * @param zeroRegion
	 *		  search datastructure for the zero region surfels
	 * @param oneRegion
	 *		  search datastructure for the one region surfels
	 * @param position
	 *		  the position at which the scale value is computed
	 * @return the scale value
	 */
	inline float computeNormalizedDistance (NeighbourHood *zeroRegion, NeighbourHood *oneRegion, Vector3D position);
	/**
	 * interpolates the scale value at <code>position</code> from the grid
	 *
	 * @param position
	 *		  the position at which the scale value is computed
	 * @return the interpolated scale value
	 */
	inline float getNormalizedDistance (Vector3D position);


private:


	float        bBox[6];       // bounding box: minx, maxx, miny, maxy, minz, maxz
	float        cellSizes[3];  // x, y, z size of individual cell
	unsigned int sizeX,
		         sizeY,
				 sizeZ;

	float        *grid;




};






inline float ScaleGrid::computeNormalizedDistance (NeighbourHood *zeroRegion,
												   NeighbourHood *oneRegion,
												   Vector3D position) {

	unsigned int i;
	unsigned int numScaleNeighbors = zeroRegion->getNofNeighbours();

	// position 
	zeroRegion->setSourcePoint (position);
	oneRegion->setSourcePoint  (position);	

	float zeroDistancePosition = 0.0f;
	float oneDistancePosition  = 0.0f;

	for (i = 0; i < numScaleNeighbors; i++) {		
		zeroDistancePosition  += sqrt (zeroRegion->getSquaredDistance (i));
		oneDistancePosition   += sqrt (oneRegion->getSquaredDistance (i));
	}
	return zeroDistancePosition / (zeroDistancePosition + oneDistancePosition);
}



inline float ScaleGrid::getNormalizedDistance (Vector3D position) {

	// tri-linear interpolation
	float weightX,
		  weightY,
		  weightZ;

	position[0] -= bBox[0];
	position[1] -= bBox[2];
	position[2] -= bBox[4];

	weightX = position[0]  / cellSizes[0];
	weightY = position[1]  / cellSizes[1];
	weightZ = position[2]  / cellSizes[2];

	unsigned int x = floor (weightX);
	unsigned int y = floor (weightY);
	unsigned int z = floor (weightZ);

	weightX -= (float) x;
	weightY -= (float) y;
	weightZ -= (float) z;

	return
		weightX * weightY * weightZ *                    grid[(x+1)*sizeY*sizeZ + (y+1)*sizeZ + z+1] + 
		weightX * weightY * (1.0f-weightZ) *             grid[(x+1)*sizeY*sizeZ + (y+1)*sizeZ + z] +
		weightX * (1 - weightY) * weightZ *              grid[(x+1)*sizeY*sizeZ + y*sizeZ + (z+1)] +
		weightX * (1 - weightY) * (1.0f-weightZ) *       grid[(x+1)*sizeY*sizeZ + y*sizeZ + z] +
		(1 - weightX) * weightY * weightZ *              grid[x*sizeY*sizeZ + (y+1)*sizeZ + z+1] + 
		(1 - weightX) * weightY * (1.0f-weightZ) *       grid[x*sizeY*sizeZ + (y+1)*sizeZ + z] +
		(1 - weightX) * (1 - weightY) * weightZ *        grid[x*sizeY*sizeZ + y*sizeZ + (z+1)] +
		(1 - weightX) * (1 - weightY) * (1.0f-weightZ) * grid[x*sizeY*sizeZ + y*sizeZ + z];

}









#endif
// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
